/*
 * Gv.c -- Gv widget.
 * Copyright (C) 1992  Timothy O. Theisen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Author: Tim Theisen           Systems Programmer
 * Internet: tim@cs.wisc.edu       Department of Computer Sciences
 *     UUCP: uwvax!tim             University of Wisconsin-Madison
 *    Phone: (608)262-0438         1210 West Dayton Street
 *      FAX: (608)262-9777         Madison, WI   53706
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Xos.h>
#include "GvP.h"
#include <ctype.h>

#ifndef XlibSpecificationRelease
typedef char *XPointer;
#endif

#include <signal.h>
#ifdef SIGNALRETURNSINT
#define SIGVAL int
#else
#define SIGVAL void
#endif

#ifdef NON_BLOCKING_IO
#include <fcntl.h>
/* if POSIX O_NONBLOCK is not available, use O_NDELAY */
#if !defined(O_NONBLOCK) && defined(O_NDELAY)
#define O_NONBLOCK O_NDELAY
#endif
#endif

#include <errno.h>
/* Both error returns are checked for non-blocking I/O. */
/* Manufacture the other error code if only one exists. */
#if !defined(EWOULDBLOCK) && defined(EAGAIN)
#define EWOULDBLOCK EAGAIN
#endif
#if !defined(EAGAIN) && defined(EWOULDBLOCK)
#define EAGAIN EWOULDBLOCK
#endif

/* GV_BUFSIZ is set to the minimum POSIX PIPE_BUF to ensure that
 * nonblocking writes to ghostscript will work properly.
 */
#define GV_BUFSIZ 512

static void ComputeXdpi(Widget w, int offset, XrmValue *value);
static void ComputeYdpi(Widget w, int offset, XrmValue *value);

static XtResource resources[] = {
#define offset(field) XtOffsetOf(GvRec, gv.field)
    { XtNarguments, XtCArguments, XtRString, sizeof(String),
	  offset(arguments), XtRString, (XtPointer)NULL },
    { XtNbottomMargin, XtCMargin, XtRInt, sizeof(int),
	  offset(bottom_margin), XtRImmediate, (XtPointer)0 },
    { XtNbusyCursor, XtCCursor, XtRCursor, sizeof(XtPointer),
	  offset(busy_cursor), XtRString, "watch" },
    { XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  offset(callback), XtRCallback, (XtPointer)NULL },
    { XtNcursor, XtCCursor, XtRCursor, sizeof(XtPointer),
	  offset(cursor), XtRString, "crosshair" },
    { XtNfilename, XtCFilename, XtRString, sizeof(String),
	  offset(filename), XtRString, (XtPointer)NULL },
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	  offset(foreground), XtRString, XtDefaultForeground},
    { XtNinterpreter, XtCInterpreter, XtRString, sizeof(String),
	  offset(interpreter), XtRString, "gs" },
    { XtNleftMargin, XtCMargin, XtRInt, sizeof(int),
	  offset(left_margin), XtRImmediate, (XtPointer)0 },
    { XtNllx, XtCBoundingBox, XtRInt, sizeof(int),
	  offset(llx), XtRImmediate, (XtPointer)0 },
    { XtNlly, XtCBoundingBox, XtRInt, sizeof(int),
	  offset(lly), XtRImmediate, (XtPointer)0 },
    { XtNmessageCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  offset(message_callback), XtRCallback, (XtPointer)NULL },
    { XtNorientation, XtCOrientation, XtRPageOrientation,
	  sizeof(XtPageOrientation), offset(orientation), XtRImmediate,
	  (XtPointer)XtPageOrientationPortrait },
    { XtNoutputCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  offset(output_callback), XtRCallback, (XtPointer)NULL },
    { XtNpalette, XtCPalette, XtRPalette, sizeof(XtPalette),
	  offset(palette), XtRImmediate, (XtPointer)XtPaletteColor },
    { XtNquiet, XtCQuiet, XtRBoolean, sizeof(Boolean),
	  offset(quiet), XtRImmediate, (XtPointer)True },
    { XtNrightMargin, XtCMargin, XtRInt, sizeof(int),
	  offset(right_margin), XtRImmediate, (XtPointer)0 },
    { XtNsafer, XtCSafer, XtRBoolean, sizeof(Boolean),
	  offset(safer), XtRImmediate, (XtPointer)True },
    { XtNtopMargin, XtCMargin, XtRInt, sizeof(int),
	  offset(top_margin), XtRImmediate, (XtPointer)0 },
    { XtNuseBackingPixmap, XtCUseBackingPixmap, XtRBoolean, sizeof(Boolean),
	  offset(use_bpixmap), XtRImmediate, (XtPointer)True },
    { XtNurx, XtCBoundingBox, XtRInt, sizeof(int),
	  offset(urx), XtRImmediate, (XtPointer)612 },
    { XtNury, XtCBoundingBox, XtRInt, sizeof(int),
	  offset(ury), XtRImmediate, (XtPointer)792 },
    { XtNxdpi, XtCResolution, XtRFloat, sizeof(float),
	  offset(xdpi), XtRCallProc, (XtPointer)ComputeXdpi },
    { XtNydpi, XtCResolution, XtRFloat, sizeof(float),
	  offset(ydpi), XtRCallProc, (XtPointer)ComputeYdpi },
#undef offset
};

static void Message(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void Notify(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void Input(XtPointer client_data, int *source, XtInputId *id);
static void Output(XtPointer client_data, int *source, XtInputId *id);

static void ClassInitialize(void);
static void ClassPartInitialize(WidgetClass class);
static void Initialize(Widget request, Widget new, ArgList args, Cardinal *num_args);
static void Realize(Widget w, Mask *valueMask, XSetWindowAttributes *attributes);
static void Destroy(Widget w);
static void Resize(Widget w);
static Boolean SetValues(Widget current, Widget request, Widget new, ArgList args, Cardinal *num_args);
static XtGeometryResult QueryGeometry(Widget w, XtWidgetGeometry *intended, XtWidgetGeometry *requested);

static void Layout(Widget w, Boolean xfree, Boolean yfree);
static Boolean ComputeSize(Widget w, Boolean xfree, Boolean yfree, Dimension *width, Dimension *height);
static void ChangeSize(Widget w, Dimension width, Dimension height);
static Boolean Setup(Widget w);
static void StartInterpreter(Widget w);
static void StopInterpreter(Widget w);
static void InterpreterFailed(Widget w);

static XtActionsRec actions[] =
{
    {"message",	Message},
    {"notify",	Notify},
};

static char translations[] =
"<Message>:		message()	\n\
";

GvClassRec gvClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &coreClassRec,
    /* class_name		*/	"Gv",
    /* widget_size		*/	sizeof(GvRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	ClassPartInitialize,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	NULL,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	QueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* gv fields */
    /* gv		*/	NULL,
    /* gv_colors		*/	NULL,
    /* next			*/	NULL,
    /* page			*/	NULL,
    /* done			*/	NULL
  }
};

WidgetClass gvWidgetClass = (WidgetClass)&gvClassRec;

/* Procedures that compute the default xdpi and ydpi from display parameters */

static void
ComputeXdpi(Widget w, int offset, XrmValue *value)
{
    static float xdpi;
    xdpi = 25.4 * WidthOfScreen(XtScreen(w)) / WidthMMOfScreen(XtScreen(w));
    value->addr = (XtPointer) &xdpi;
}

static void
ComputeYdpi(Widget w, int offset, XrmValue *value)
{
    static float ydpi;
    ydpi = 25.4 * HeightOfScreen(XtScreen(w)) / HeightMMOfScreen(XtScreen(w));
    value->addr = (XtPointer) &ydpi;
}

/* Message action routine.
 * Passes ghostscript message events back to application via
 * the message callback.  It also marks the interpreter as
 * being not busy at the end of page, and stops the interpreter
 * when it send a "done" message.
 */
static void
Message(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    GvWidget gvw = (GvWidget) w;
    GvWidgetClass gvc = (GvWidgetClass) XtClass(w);

    gvw->gv.mwin = event->xclient.data.l[0];
    if (event->xclient.message_type ==
	XmuInternAtom(XtDisplay(w), gvc->gv_class.page)) {
	gvw->gv.busy = False;
	XDefineCursor(XtDisplay(gvw), XtWindow(gvw), gvw->gv.cursor);
	XtCallCallbackList(w, gvw->gv.message_callback, "Page");
    } else if (event->xclient.message_type ==
	       XmuInternAtom(XtDisplay(w), gvc->gv_class.done)) {
	StopInterpreter(w);
	XtCallCallbackList(w, gvw->gv.message_callback, "Done");
    }
}

/* Notify action routine.
 * Calculates where the user clicked in the default user coordinate system.
 * Call the callbacks with the point of click and size of zoom window
 * requested.
 */
static void
Notify(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    GvWidget gvw = (GvWidget) w;

    XtCallCallbackList(w, gvw->gv.callback, (XtPointer) NULL);
}

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

static Boolean broken_pipe = False;

static SIGVAL
CatchPipe(int i)
{
    broken_pipe = True;
#ifdef SIGNALRETURNSINT
    return 0;
#endif
}

/* Input - Feed data to ghostscript's stdin.
 * Write bytes to ghostscript using non-blocking I/O.
 * Also, pipe signals are caught during writing.  The return
 * values are checked and the appropriate action is taken.  I do
 * this at this low level, because it may not be appropriate for
 * SIGPIPE to be caught for the overall application.
 */

static void
Input(XtPointer client_data, int *source, XtInputId *id)
{
    Widget w = (Widget) client_data;
    GvWidget gvw = (GvWidget) w;
    int bytes_written;
    SIGVAL (*oldsig)(int);

    oldsig = signal(SIGPIPE, CatchPipe);

#ifdef NON_BLOCKING_IO
    do {
#endif

	if (gvw->gv.buffer_bytes_left == 0) {

	    /* Get a new section if required */
	    if (gvw->gv.ps_input && gvw->gv.bytes_left == 0) {
		struct record_list *ps_old = gvw->gv.ps_input;
		gvw->gv.ps_input = ps_old->next;
		if (ps_old->close) fclose(ps_old->fp);
		XtFree((char *)ps_old);
	    }

	    /* Have to seek at the beginning of each section */
	    if (gvw->gv.ps_input &&
		gvw->gv.ps_input->seek_needed) {
		if (gvw->gv.ps_input->len > 0)
		    fseek(gvw->gv.ps_input->fp,
			  gvw->gv.ps_input->begin, SEEK_SET);
		gvw->gv.ps_input->seek_needed = False;
		gvw->gv.bytes_left = gvw->gv.ps_input->len;
	    }

	    if (gvw->gv.bytes_left > GV_BUFSIZ) {
		gvw->gv.buffer_bytes_left =
			fread(gvw->gv.input_buffer,
			      sizeof (char), GV_BUFSIZ,
			      gvw->gv.ps_input->fp);
	    } else if (gvw->gv.bytes_left > 0) {
		gvw->gv.buffer_bytes_left =
			fread(gvw->gv.input_buffer,
			      sizeof (char), gvw->gv.bytes_left,
			      gvw->gv.ps_input->fp);
	    } else {
		gvw->gv.buffer_bytes_left = 0;
	    }
	    if (gvw->gv.bytes_left > 0 &&
		gvw->gv.buffer_bytes_left == 0) {
		InterpreterFailed(w);	/* Error occurred */
	    }
	    gvw->gv.input_buffer_ptr = gvw->gv.input_buffer;
	    gvw->gv.bytes_left -= gvw->gv.buffer_bytes_left;
	}

	if (gvw->gv.buffer_bytes_left > 0) {
	    bytes_written = write(gvw->gv.interpreter_input,
				  gvw->gv.input_buffer_ptr,
				  gvw->gv.buffer_bytes_left);

	    if (broken_pipe) {
		broken_pipe = False;
		InterpreterFailed(w);		/* Something bad happened */
	    } else if (bytes_written == -1) {
		if ((errno != EWOULDBLOCK) && (errno != EAGAIN)) {
		    InterpreterFailed(w);	/* Something bad happened */
		}
	    } else {
		gvw->gv.buffer_bytes_left -= bytes_written;
		gvw->gv.input_buffer_ptr += bytes_written;
	    }
	}
#ifdef NON_BLOCKING_IO
    } while(gvw->gv.ps_input &&
	    gvw->gv.buffer_bytes_left == 0);
#endif
    signal(SIGPIPE, oldsig);
    if (gvw->gv.ps_input == NULL &&
	gvw->gv.buffer_bytes_left == 0) {
	if (gvw->gv.interpreter_input_id != None) {
	    XtRemoveInput(gvw->gv.interpreter_input_id);
	    gvw->gv.interpreter_input_id = None;
	}
    }
}

/* Output - receive I/O from ghostscript's stdout and stderr.
 * Pass this to the application via the output_callback. */
static void
Output(XtPointer client_data, int *source, XtInputId *id)
{
    Widget w = (Widget) client_data;
    GvWidget gvw = (GvWidget) w;
    char buf[GV_BUFSIZ+1];
    int bytes = 0;

    if (*source == gvw->gv.interpreter_output) {
	bytes = read(gvw->gv.interpreter_output, buf, GV_BUFSIZ);
	if (bytes == 0) { /* EOF occurred */
	    close(gvw->gv.interpreter_output);
	    gvw->gv.interpreter_output = -1;
	    XtRemoveInput(gvw->gv.interpreter_output_id);
	    return;
	} else if (bytes == -1) {
	    InterpreterFailed(w);		/* Something bad happened */
	    return;
	}
    } else if (*source == gvw->gv.interpreter_error) {
	bytes = read(gvw->gv.interpreter_error, buf, GV_BUFSIZ);
	if (bytes == 0) { /* EOF occurred */
	    close(gvw->gv.interpreter_error);
	    gvw->gv.interpreter_error = -1;
	    XtRemoveInput(gvw->gv.interpreter_error_id);
	    return;
	} else if (bytes == -1) {
	    InterpreterFailed(w);		/* Something bad happened */
	    return;
	}
    }
    if (bytes > 0) {
	buf[bytes] = '\0';
	XtCallCallbackList(w, gvw->gv.output_callback, (XtPointer) buf);
    }
}

/* Register the type converter required for the PageOrientation. */
/* Register the type converter required for the Palette. */
/* This routine is called exactly once. */
static void
ClassInitialize(void)
{
    XtSetTypeConverter(XtRString, XtRPageOrientation,
		       XmuCvtStringToPageOrientation, NULL, 0,
		       XtCacheAll, NULL);
    XtSetTypeConverter(XtRString, XtRPalette,
		       XmuCvtStringToPalette, NULL, 0,
		       XtCacheAll, NULL);
}

/* Get atoms needed to communicate with ghostscript. */
/* This routine is called once per display. */
static void
ClassPartInitialize(WidgetClass class)
{
    GvWidgetClass gvc = (GvWidgetClass)class;
    gvc->gv_class.gv = XmuMakeAtom("GHOSTVIEW");
    gvc->gv_class.gv_colors = XmuMakeAtom("GHOSTVIEW_COLORS");
    gvc->gv_class.next = XmuMakeAtom("NEXT");
    gvc->gv_class.page = XmuMakeAtom("PAGE");
    gvc->gv_class.done = XmuMakeAtom("DONE");
}

/* Initialize private state. */

static void
Initialize(Widget request, Widget new, ArgList args, Cardinal *num_args)
{
    XGCValues	values;
    XtGCMask	mask;
    GvWidget ngvw = (GvWidget) new;
    GvWidget rgvw = (GvWidget) request;

    values.foreground = new->core.background_pixel;
    mask = GCForeground;
    ngvw->gv.gc = XtGetGC(new, mask, &values);
    ngvw->gv.mwin = None;
    ngvw->gv.disable_start = False;
    ngvw->gv.interpreter_pid = -1;
    ngvw->gv.input_buffer = NULL;
    ngvw->gv.bytes_left = 0;
    ngvw->gv.input_buffer_ptr = NULL;
    ngvw->gv.buffer_bytes_left = 0;
    ngvw->gv.ps_input = NULL;
    ngvw->gv.interpreter_input = -1;
    ngvw->gv.interpreter_output = -1;
    ngvw->gv.interpreter_error = -1;
    ngvw->gv.interpreter_input_id = None;
    ngvw->gv.interpreter_output_id = None;
    ngvw->gv.interpreter_error_id = None;
    ngvw->gv.gs_width = 0;
    ngvw->gv.gs_height = 0;
    ngvw->gv.changed = False;
    ngvw->gv.busy = False;

    /* Compute window size */
    Layout(new, (rgvw->core.width == 0), (rgvw->core.height == 0));
}

/* Create Window and start interpreter if needed */
static void
Realize(Widget w, Mask *valueMask, XSetWindowAttributes *attributes)
{
    GvWidget gvw = (GvWidget) w;

    if (gvw->gv.cursor != None) {
	attributes->cursor = gvw->gv.cursor;
	*valueMask |= CWCursor;
    }

    XtCreateWindow(w, (unsigned int) InputOutput, (Visual *) CopyFromParent,
		   *valueMask, attributes);

    Setup(w);
}

/* Destroy routine: kill the interpreter and release the GC */
static void
Destroy(Widget w)
{
    GvWidget gvw = (GvWidget) w;

    StopInterpreter(w);
    XtReleaseGC(w, gvw->gv.gc);
    if (gvw->gv.input_buffer) XtFree(gvw->gv.input_buffer);
    if (gvw->core.background_pixmap != XtUnspecifiedPixmap)
	XFreePixmap(XtDisplay(w), gvw->core.background_pixmap);
}

/* Process resize request.  Requested size cannot be changed.
 * NOTE: This routine may be called before the widget is realized.
 * (It was a surprise to me.)
 * If the widget is realized, start a new interpreter by calling Setup().
 * If Setup() actually started a new interpreter and it is taking input
 * from stdin, send a refresh message to the application.  This is the
 * only way that the application can be notified that it needs to resend
 * the input because someone forced a new window size on the widget.
 */
static void
Resize(Widget w)
{
    Layout(w, False, False);
    if (!XtIsRealized(w)) return;
    if (Setup(w)) {
	GvWidget gvw = (GvWidget) w;
	if (gvw->gv.filename == NULL) {
	    XtCallCallbackList(w, gvw->gv.message_callback, "Refresh");
	}
    }
}

/* SetValues routine.  Set new private state, based on changed values
 * in the widget.  Always returns False, because redisplay is never needed.
 */
static Boolean
SetValues(Widget current, Widget request, Widget new, ArgList args, Cardinal *num_args)
{
    GvWidget cgvw = (GvWidget) current;
    GvWidget rgvw = (GvWidget) request;
    GvWidget ngvw = (GvWidget) new;
    String cfilename;
    String rfilename;
    String carguments;
    String rarguments;

    cfilename = cgvw->gv.filename;
    if (cfilename == NULL) cfilename = "(null)";
    rfilename = rgvw->gv.filename;
    if (rfilename == NULL) rfilename = "(null)";
    carguments = cgvw->gv.arguments;
    if (carguments == NULL) carguments = "(null)";
    rarguments = rgvw->gv.arguments;
    if (rarguments == NULL) rarguments = "(null)";

    if (XtIsRealized(new) && !ngvw->gv.busy &&
	(cgvw->gv.cursor != ngvw->gv.cursor)) {
	XDefineCursor(XtDisplay(new), XtWindow(new), ngvw->gv.cursor);
    }
    if (XtIsRealized(new) && ngvw->gv.busy &&
	(cgvw->gv.busy_cursor != ngvw->gv.busy_cursor)) {
	XDefineCursor(XtDisplay(new), XtWindow(new),
		      ngvw->gv.busy_cursor);
    }
    if (cgvw->core.background_pixel != rgvw->core.background_pixel) {
	XGCValues	values;
	XtGCMask	mask;

	XtReleaseGC(current, cgvw->gv.gc);
	values.foreground = new->core.background_pixel;
	mask = GCForeground;
	ngvw->gv.gc = XtGetGC(new, mask, &values);
    }
    if ((cgvw->core.width != rgvw->core.width) ||
	(cgvw->core.height != rgvw->core.height) ||
	(cgvw->core.background_pixel != rgvw->core.background_pixel) ||
	(cgvw->gv.foreground != rgvw->gv.foreground) ||
	(cgvw->gv.palette != rgvw->gv.palette) ||
	strcmp(cgvw->gv.interpreter, rgvw->gv.interpreter) ||
	strcmp(carguments, rarguments) ||
	(cgvw->gv.quiet != rgvw->gv.quiet) ||
	(cgvw->gv.safer != rgvw->gv.safer) ||
	strcmp(cfilename, rfilename) ||
	(cgvw->gv.orientation != rgvw->gv.orientation) ||
	(cgvw->gv.use_bpixmap != rgvw->gv.use_bpixmap) ||
	(cgvw->gv.xdpi != rgvw->gv.xdpi) ||
	(cgvw->gv.ydpi != rgvw->gv.ydpi) ||
	(cgvw->gv.bottom_margin != rgvw->gv.bottom_margin) ||
	(cgvw->gv.left_margin != rgvw->gv.left_margin) ||
	(cgvw->gv.right_margin != rgvw->gv.right_margin) ||
	(cgvw->gv.top_margin != rgvw->gv.top_margin) ||
	(cgvw->gv.llx != rgvw->gv.llx) ||
	(cgvw->gv.lly != rgvw->gv.lly) ||
	(cgvw->gv.urx != rgvw->gv.urx) ||
	(cgvw->gv.ury != rgvw->gv.ury)) {

	ngvw->gv.changed = True;
	Layout(new, True, True);
    }

    if (ngvw->gv.changed && XtIsRealized(current)) Setup(new);
    return(False);
}

/*	Function Name: QueryGeometry
 *	Description: This tells the parent what size we would like to be
 *                   given certain constraints.
 *	Arguments: w - the widget.
 *                 intended - what the parent intends to do with us.
 *                 requested - what we want to happen.
 */

static XtGeometryResult 
QueryGeometry(Widget w, XtWidgetGeometry *intended, XtWidgetGeometry *requested)
{
    Dimension new_width, new_height;
    Boolean change, width_req, height_req;
    
    width_req = intended->request_mode & CWWidth;
    height_req = intended->request_mode & CWHeight;

    if (width_req)
      new_width = intended->width;
    else
      new_width = w->core.width;

    if (height_req)
      new_height = intended->height;
    else
      new_height = w->core.height;

    requested->request_mode = 0;
    
/*
 * We only care about our height and width.
 */

    if (!width_req && !height_req)
      return(XtGeometryYes);
    
    change = ComputeSize(w, !width_req, !height_req, &new_width, &new_height);

    requested->request_mode |= CWWidth;
    requested->width = new_width;
    requested->request_mode |= CWHeight;
    requested->height = new_height;

    if (change)
        return(XtGeometryAlmost);
    return(XtGeometryYes);
}

/* Layout the widget. */

static void
Layout(Widget w, Boolean xfree, Boolean yfree)
{
    Dimension width = w->core.width;
    Dimension height = w->core.height;
    Boolean different_size = ComputeSize(w, xfree, yfree, &width, &height);
    if (different_size) ChangeSize(w, width, height);
}

/* Compute new size of window, sets xdpi and ydpi if necessary.
 * returns True if new window size is different */
static Boolean
ComputeSize(Widget w, Boolean xfree, Boolean yfree, Dimension *width, Dimension *height)
{
    GvWidget gvw = (GvWidget) w;
    Dimension new_width = *width;
    Dimension new_height = *height;
    float newxdpi, newydpi;
    Boolean change;

    if (xfree && yfree) {
	/* width and height can be changed, calculate window size according */
	/* to xpdi and ydpi */
	switch (gvw->gv.orientation) {
	case XtPageOrientationPortrait:
	case XtPageOrientationUpsideDown:
	    new_width = (gvw->gv.urx - gvw->gv.llx) / 72.0 *
			 gvw->gv.xdpi + 0.5;
	    new_height = (gvw->gv.ury - gvw->gv.lly) / 72.0 *
			  gvw->gv.ydpi + 0.5;
	    break;
	case XtPageOrientationLandscape:
	case XtPageOrientationSeascape:
	    new_width = (gvw->gv.ury - gvw->gv.lly) / 72.0 *
			 gvw->gv.xdpi + 0.5;
	    new_height = (gvw->gv.urx - gvw->gv.llx) / 72.0 *
			  gvw->gv.ydpi + 0.5;
	    break;
	}
    } else if (xfree) {
	/* height is fixed.  Preserve aspect ratio by recomputing */
	/* ydpi and xdpi */
	switch (gvw->gv.orientation) {
	case XtPageOrientationPortrait:
	case XtPageOrientationUpsideDown:
	    newydpi = gvw->core.height * 72.0 /
		      (gvw->gv.ury - gvw->gv.lly);
	    newxdpi = newydpi * gvw->gv.xdpi / gvw->gv.ydpi;
	    gvw->gv.xdpi = newxdpi;
	    gvw->gv.ydpi = newydpi;
	    new_width = (gvw->gv.urx - gvw->gv.llx) / 72.0 *
			 gvw->gv.xdpi + 0.5;
	    break;
	case XtPageOrientationLandscape:
	case XtPageOrientationSeascape:
	    newydpi = gvw->core.height * 72.0 /
		      (gvw->gv.urx - gvw->gv.llx);
	    newxdpi = newydpi * gvw->gv.xdpi / gvw->gv.ydpi;
	    gvw->gv.xdpi = newxdpi;
	    gvw->gv.ydpi = newydpi;
	    new_width = (gvw->gv.ury - gvw->gv.lly) / 72.0 *
			 gvw->gv.xdpi + 0.5;
	    break;
	}
    } else if (yfree) {
	/* width is fixed.  Preserve aspect ratio by recomputing */
	/* xdpi and ydpi */
	switch (gvw->gv.orientation) {
	case XtPageOrientationPortrait:
	case XtPageOrientationUpsideDown:
	    newxdpi = gvw->core.width * 72.0 /
		      (gvw->gv.urx - gvw->gv.llx);
	    newydpi = newxdpi * gvw->gv.ydpi / gvw->gv.xdpi;
	    gvw->gv.xdpi = newxdpi;
	    gvw->gv.ydpi = newydpi;
	    new_height = (gvw->gv.ury - gvw->gv.lly) / 72.0 *
			  gvw->gv.ydpi + 0.5;
	    break;
	case XtPageOrientationLandscape:
	case XtPageOrientationSeascape:
	    newxdpi = gvw->core.width * 72.0 /
		      (gvw->gv.ury - gvw->gv.lly);
	    newydpi = newxdpi * gvw->gv.ydpi / gvw->gv.xdpi;
	    gvw->gv.xdpi = newxdpi;
	    gvw->gv.ydpi = newydpi;
	    new_height = (gvw->gv.urx - gvw->gv.llx) / 72.0 *
			  gvw->gv.ydpi + 0.5;
	    break;
	}
    } else {
	/* height and width are fixed.  Just have to live with it. */
	switch (gvw->gv.orientation) {
	case XtPageOrientationPortrait:
	case XtPageOrientationUpsideDown:
	    gvw->gv.xdpi = gvw->core.width * 72.0 /
				  (gvw->gv.urx - gvw->gv.llx);
	    gvw->gv.ydpi = gvw->core.height * 72.0 /
				  (gvw->gv.ury - gvw->gv.lly);
	    break;
	case XtPageOrientationLandscape:
	case XtPageOrientationSeascape:
	    gvw->gv.xdpi = gvw->core.width * 72.0 /
				  (gvw->gv.ury - gvw->gv.lly);
	    gvw->gv.ydpi = gvw->core.height * 72.0 /
				  (gvw->gv.urx - gvw->gv.llx);
	    break;
	}
    }

    change = (new_width != *width) || (new_height != *height);
    *width = new_width;
    *height = new_height;
    return (change);
}

/*	Function Name: ChangeSize.
 *	Description: Request a size change.
 *	Arguments: w - the widget to try change the size of.
 */

static void
ChangeSize(Widget w, Dimension width, Dimension height)
{
    XtWidgetGeometry request, reply;
    Boolean changed = False;

    request.request_mode = CWWidth | CWHeight;
    request.width = width;
    request.height = height;
    
    switch ( XtMakeGeometryRequest(w, &request, &reply) ) {
    case XtGeometryYes:
	changed = True;
        break;
    case XtGeometryNo:
        break;
    case XtGeometryAlmost:
	ComputeSize(w, (request.height != reply.height),
		       (request.width != reply.width),
		       &(reply.width), &(reply.height));
	request = reply;
	switch (XtMakeGeometryRequest(w, &request, &reply) ) {
	case XtGeometryYes:
	    changed = True;
	    break;
	case XtGeometryNo:
	    break;
	case XtGeometryAlmost:
	    request = reply;
	    ComputeSize(w, FALSE, FALSE, &(request.width), &(request.height));
	    request.request_mode = CWWidth | CWHeight;
	    XtMakeGeometryRequest(w, &request, &reply);
	    changed = True;
	    break;
	default:
	    break;
	}
	break;
    default:
	break;
    }

    /* If success, setup the widet for the new size. */
    if (changed && XtIsRealized(w)) Setup(w);
}

/* Catch the alloc error when there is not enough resources for the
 * backing pixmap.  Automatically shut off backing pixmap and let the
 * user know when this happens.
 */
static Boolean alloc_error;
static XErrorHandler oldhandler;

static int
catch_alloc(Display *dpy, XErrorEvent *err)
{
    if (err->error_code == BadAlloc) {
	alloc_error = True;
    }
    if (alloc_error) return 0;
    return oldhandler(dpy, err);
}

/* Setup - sets up the backing pixmap, and GHOSTVIEW property and
 * starts interpreter if needed.
 * NOTE: the widget must be realized before calling Setup().
 * Returns True if a new interpreter was started, False otherwise.
 */
 
static Boolean
Setup(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    GvWidgetClass gvc = (GvWidgetClass) XtClass(w);
    char buf[GV_BUFSIZ];
    Pixmap bpixmap;
    XSetWindowAttributes xswa;

    if (!gvw->gv.changed &&
	(gvw->core.width == gvw->gv.gs_width) &&
	(gvw->core.height == gvw->gv.gs_height)) return False;

    StopInterpreter(w);

    if ((gvw->core.width != gvw->gv.gs_width) ||
	(gvw->core.height != gvw->gv.gs_height) ||
	(!gvw->gv.use_bpixmap)) {
	if (gvw->core.background_pixmap != XtUnspecifiedPixmap) {
	    XFreePixmap(XtDisplay(w), gvw->core.background_pixmap);
	    gvw->core.background_pixmap = XtUnspecifiedPixmap;
	    XSetWindowBackgroundPixmap(XtDisplay(w), XtWindow(w), None);
	}
    }

    if (gvw->gv.use_bpixmap) {
	if (gvw->core.background_pixmap == XtUnspecifiedPixmap) {
	    /* Get a Backing Pixmap, but be ready for the BadAlloc. */
	    XSync(XtDisplay(w), False);  /* Get to known state */
	    oldhandler = XSetErrorHandler(catch_alloc);
	    alloc_error = False;
	    bpixmap = XCreatePixmap(XtDisplay(w), XtWindow(w),
				    gvw->core.width, gvw->core.height,
				    gvw->core.depth);
	    XSync(XtDisplay(w), False);  /* Force the error */
	    if (alloc_error) {
		XtCallCallbackList(w, gvw->gv.message_callback,
				   "BadAlloc");
		if (bpixmap != None) {
		    XFreePixmap(XtDisplay(w), bpixmap);
		    XSync(XtDisplay(w), False);  /* Force the error */
		    bpixmap = None;
		}
	    }
	    oldhandler = XSetErrorHandler(oldhandler);
	    if (bpixmap != None) {
		gvw->core.background_pixmap = bpixmap;
		XSetWindowBackgroundPixmap(XtDisplay(w), XtWindow(w),
					   gvw->core.background_pixmap);
	    }
	} else {
	    bpixmap = gvw->core.background_pixmap;
	}
    } else {
	if (gvw->core.background_pixmap != XtUnspecifiedPixmap) {
	    XFreePixmap(XtDisplay(w), gvw->core.background_pixmap);
	    gvw->core.background_pixmap = XtUnspecifiedPixmap;
	    XSetWindowBackgroundPixmap(XtDisplay(w), XtWindow(w), None);
	}
	bpixmap = None;
    }

    if (bpixmap != None) {
	xswa.backing_store = NotUseful;
	XChangeWindowAttributes(XtDisplay(w), XtWindow(w),
				CWBackingStore, &xswa);
    } else {
	xswa.backing_store = Always;
	XChangeWindowAttributes(XtDisplay(w), XtWindow(w),
				CWBackingStore, &xswa);
    }

    gvw->gv.gs_width = gvw->core.width;
    gvw->gv.gs_height = gvw->core.height;

    sprintf(buf, "%ld %d %d %d %d %d %g %g %d %d %d %d",
	    bpixmap, gvw->gv.orientation,
	    gvw->gv.llx, gvw->gv.lly,
	    gvw->gv.urx, gvw->gv.ury,
	    gvw->gv.xdpi, gvw->gv.ydpi,
	    gvw->gv.left_margin, gvw->gv.bottom_margin,
	    gvw->gv.right_margin, gvw->gv.top_margin);
    XChangeProperty(XtDisplay(w), XtWindow(w),
		   XmuInternAtom(XtDisplay(w), gvc->gv_class.gv),
		   XA_STRING, 8, PropModeReplace,
		   (unsigned char *)buf, strlen(buf));

    sprintf(buf, "%s %ld %ld",
	    gvw->gv.palette == XtPaletteMonochrome ? "Monochrome" :
	    gvw->gv.palette == XtPaletteGrayscale  ? "Grayscale" :
	    gvw->gv.palette == XtPaletteColor      ? "Color" : "?",
	    gvw->gv.foreground, gvw->core.background_pixel);
    XChangeProperty(XtDisplay(w), XtWindow(w),
		   XmuInternAtom(XtDisplay(w), gvc->gv_class.gv_colors),
		   XA_STRING, 8, PropModeReplace,
		   (unsigned char *)buf, strlen(buf));

    XSync(XtDisplay(w), False);  /* Be sure to update properties */
    StartInterpreter(w);
    return True;
}

/* This routine starts the interpreter.  It sets the DISPLAY and 
 * GHOSTVIEW environment variables.  The GHOSTVIEW environment variable
 * contains the Window that ghostscript should write on.
 *
 * This routine also opens pipes for stdout and stderr and initializes
 * application input events for them.  If input to ghostscript is not
 * from a file, a pipe for stdin is created.  This pipe is setup for
 * non-blocking I/O so that the user interface never "hangs" because of
 * a write to ghostscript.
 */
static void
StartInterpreter(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    int	std_in[2];
    int	std_out[2];
    int	std_err[2];
    char buf[GV_BUFSIZ];
#define NUM_ARGS 100
    char *argv[NUM_ARGS];
    char *arguments = NULL;
    char *cptr;
    int argc = 0;
    int ret;

    StopInterpreter(w);

    /* Clear the window before starting a new interpreter. */
    if (gvw->core.background_pixmap != XtUnspecifiedPixmap) {
	XFillRectangle(XtDisplay(w), gvw->core.background_pixmap,
		       gvw->gv.gc,
		       0, 0, gvw->core.width, gvw->core.height);
    }
    XClearArea(XtDisplay(w), XtWindow(w),
	       0, 0, gvw->core.width, gvw->core.height, False);

    if (gvw->gv.disable_start) return;

    argv[argc++] = gvw->gv.interpreter;
    argv[argc++] = "-sDEVICE=x11";
    argv[argc++] = "-dNOPAUSE";
    if (gvw->gv.quiet) argv[argc++] = "-dQUIET";
    if (gvw->gv.safer) argv[argc++] = "-dSAFER";
    if (gvw->gv.arguments) {
	cptr = arguments = XtNewString(gvw->gv.arguments);
	while (isspace(*cptr)) cptr++;
	while (*cptr) {
	    argv[argc++] = cptr;
	    while (*cptr && !isspace(*cptr)) cptr++;
	    if (*cptr) *cptr++ = '\0';
	    if (argc + 2 >= NUM_ARGS) {
		fprintf(stderr, "Too many arguments to interpreter.\n");
		exit(1);
	    }
	    while (isspace(*cptr)) cptr++;
	}
    }
    argv[argc++] = "-";
    argv[argc++] = NULL;

    if (gvw->gv.filename == NULL) {
	ret = pipe(std_in);
	if (ret == -1) {
	    perror("Could not create pipe");
	    exit(1);
	}
    } else if (strcmp(gvw->gv.filename, "-")) {
	std_in[0] = open(gvw->gv.filename, O_RDONLY, 0);
    }
    ret = pipe(std_out);
    if (ret == -1) {
	perror("Could not create pipe");
	exit(1);
    }
    ret = pipe(std_err);
    if (ret == -1) {
	perror("Could not create pipe");
	exit(1);
    }

    gvw->gv.changed = False;
    gvw->gv.busy = True;
    XDefineCursor(XtDisplay(gvw), XtWindow(gvw), gvw->gv.busy_cursor);
    gvw->gv.interpreter_pid = fork();

    if (gvw->gv.interpreter_pid == 0) { /* child */
	close(std_out[0]);
	close(std_err[0]);
	dup2(std_out[1], 1);
	close(std_out[1]);
	dup2(std_err[1], 2);
	close(std_err[1]);
	sprintf(buf, "%ld", XtWindow(w));
	setenv("GHOSTVIEW", buf, True);
	setenv("DISPLAY", XDisplayString(XtDisplay(w)), True);
	if (gvw->gv.filename == NULL) {
	    close(std_in[1]);
	    dup2(std_in[0], 0);
	    close(std_in[0]);
	} else if (strcmp(gvw->gv.filename, "-")) {
	    dup2(std_in[0], 0);
	    close(std_in[0]);
	}
	execvp(argv[0], argv);
	sprintf(buf, "Exec of %s failed", argv[0]);
	perror(buf);
	_exit(1);
    } else {
	if (gvw->gv.filename == NULL) {
#ifdef NON_BLOCKING_IO
	    int result;
#endif
	    close(std_in[0]);

#ifdef NON_BLOCKING_IO
	    result = fcntl(std_in[1], F_GETFL, 0);
	    result = result | O_NONBLOCK;
	    result = fcntl(std_in[1], F_SETFL, result);
#endif
	    gvw->gv.interpreter_input = std_in[1];
	    gvw->gv.interpreter_input_id = None;
	} else if (strcmp(gvw->gv.filename, "-")) {
	    close(std_in[0]);
	}
	close(std_out[1]);
	gvw->gv.interpreter_output = std_out[0];
	gvw->gv.interpreter_output_id = 
	    XtAppAddInput(XtWidgetToApplicationContext(w), std_out[0],
			  (XtPointer)XtInputReadMask, Output, (XtPointer)w);
	close(std_err[1]);
	gvw->gv.interpreter_error = std_err[0];
	gvw->gv.interpreter_error_id = 
	    XtAppAddInput(XtWidgetToApplicationContext(w), std_err[0],
			  (XtPointer)XtInputReadMask, Output, (XtPointer)w);
    }
    if (arguments) XtFree(arguments);
}

/* Stop the interperter, if present, and remove any Input sources. */
/* Also reset the busy state. */
static void
StopInterpreter(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    if (gvw->gv.interpreter_pid >= 0) {
	kill(gvw->gv.interpreter_pid, SIGTERM);
	wait(0);
	gvw->gv.interpreter_pid = -1;
    }
    if (gvw->gv.interpreter_input >= 0) {
	close(gvw->gv.interpreter_input);
	gvw->gv.interpreter_input = -1;
	if (gvw->gv.interpreter_input_id != None) {
	    XtRemoveInput(gvw->gv.interpreter_input_id);
	    gvw->gv.interpreter_input_id = None;
	}
	while (gvw->gv.ps_input) {
	    struct record_list *ps_old = gvw->gv.ps_input;
	    gvw->gv.ps_input = ps_old->next;
	    if (ps_old->close) fclose(ps_old->fp);
	    XtFree((char *)ps_old);
	}
    }
    if (gvw->gv.interpreter_output >= 0) {
	close(gvw->gv.interpreter_output);
	gvw->gv.interpreter_output = -1;
	XtRemoveInput(gvw->gv.interpreter_output_id);
    }
    if (gvw->gv.interpreter_error >= 0) {
	close(gvw->gv.interpreter_error);
	gvw->gv.interpreter_error = -1;
	XtRemoveInput(gvw->gv.interpreter_error_id);
    }
    gvw->gv.busy = False;
    XDefineCursor(XtDisplay(gvw), XtWindow(gvw), gvw->gv.cursor);
}

/* The interpeter failed, Stop what's left and notify application */
static void
InterpreterFailed(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    StopInterpreter(w);
    XtCallCallbackList(w, gvw->gv.message_callback, "Failed");
}

/*
 *	Public Routines
 */

/* GvDisableInterpreter:
 * Stop any interpreter and disable new ones from starting.
 */
void
GvDisableInterpreter(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    gvw->gv.disable_start = True;
    if (XtIsRealized(w)) StopInterpreter(w);
}

/* GvDisableInterpreter:
 * Allow an interpreter to start and start one if the widget is
 * currently realized.
 */
void
GvEnableInterpreter(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    gvw->gv.disable_start = False;
    if (XtIsRealized(w)) StartInterpreter(w);
}

/* GvIsInterpreterReady:
 * Returns True if the interpreter is ready for new input.
 */
Boolean
GvIsInterpreterReady(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    return gvw->gv.interpreter_pid != -1 &&
	   !gvw->gv.busy &&
	   gvw->gv.ps_input == NULL;
}

/* GvIsInterpreterRunning:
 * Returns True if the interpreter is running.
 */
Boolean
GvIsInterpreterRunning(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    return gvw->gv.interpreter_pid != -1;
}

/* GvGetBackingPixmap:
 * Returns the current backing pixmap.
 */
Pixmap
GvGetBackingPixmap(Widget w)
{
    GvWidget gvw = (GvWidget) w;
    if (gvw->core.background_pixmap != XtUnspecifiedPixmap)
	return(gvw->core.background_pixmap);
    else
	return(None);
}

/* GvSendPS:
 *   Queue a portion of a PostScript file for output to ghostscript.
 *   fp: FILE * of the file in question.  NOTE: if you have several
 *   Gv widgets reading from the same file.  You must open
 *   a unique FILE * for each widget.
 *   SendPS does not actually send the PostScript, it merely queues it
 *   for output.
 *   begin: position in file (returned from ftell()) to start.
 *   len:   number of bytes to write.
 *
 *   If an interpreter is not running, nothing is queued and
 *   False is returned.
 */
Boolean
GvSendPS(w, fp, begin, len, close)
    Widget w;
    FILE *fp;
    long begin;
    unsigned int len;
    Bool close;
{
    GvWidget gvw = (GvWidget) w;
    struct record_list *ps_new;

    if (gvw->gv.interpreter_input < 0) return False;
    ps_new = (struct record_list *) XtMalloc(sizeof (struct record_list));
    ps_new->fp = fp;
    ps_new->begin = begin;
    ps_new->len = len;
    ps_new->seek_needed = True;
    ps_new->close = close;
    ps_new->next = NULL;

    if (gvw->gv.input_buffer == NULL) {
	gvw->gv.input_buffer = XtMalloc(GV_BUFSIZ);
    }

    if (gvw->gv.ps_input == NULL) {
	gvw->gv.input_buffer_ptr = gvw->gv.input_buffer;
	gvw->gv.bytes_left = len;
	gvw->gv.buffer_bytes_left = 0;
	gvw->gv.ps_input = ps_new;
	gvw->gv.interpreter_input_id =
		XtAppAddInput(XtWidgetToApplicationContext(w),
			      gvw->gv.interpreter_input,
			      (XtPointer)XtInputWriteMask, Input, (XtPointer)w);
    } else {
	struct record_list *p = gvw->gv.ps_input;
	while (p->next != NULL) {
	    p = p->next;
	}
	p->next = ps_new;
    }
    return True;
}

/* GvNextPage:
 *   Tell ghostscript to start the next page.
 *   Returns False if ghostscript is not running, or not ready to start
 *   another page.
 *   If another page is started.  Sets the busy flag and cursor.
 */
Boolean
GvNextPage(w)
    Widget w;
{
    GvWidget gvw = (GvWidget) w;
    GvWidgetClass gvc = (GvWidgetClass) XtClass(w);
    XEvent event;

    if (gvw->gv.interpreter_pid < 0) return False;
    if (gvw->gv.mwin == None) return False;

    if (!gvw->gv.busy) {
	gvw->gv.busy = True;
	XDefineCursor(XtDisplay(gvw), XtWindow(gvw),
		      gvw->gv.busy_cursor);

	event.xclient.type = ClientMessage;
	event.xclient.display = XtDisplay(w);
	event.xclient.window = gvw->gv.mwin;
	event.xclient.message_type =
		XmuInternAtom(XtDisplay(w), gvc->gv_class.next);
	event.xclient.format = 32;
	XSendEvent(XtDisplay(w), gvw->gv.mwin, False, 0, &event);
	XFlush(XtDisplay(w));	/* And push it out */
	return True;
    } else {
	return False;
    }
}

#define	done(type, value) \
	{							\
	    if (toVal->addr != NULL) {				\
		if (toVal->size < sizeof(type)) {		\
		    toVal->size = sizeof(type);			\
		    return False;				\
		}						\
		*(type*)(toVal->addr) = (value);		\
	    }							\
	    else {						\
		static type static_val;				\
		static_val = (value);				\
		toVal->addr = (XPointer)&static_val;		\
	    }							\
	    toVal->size = sizeof(type);				\
	    return True;					\
	}

/* PageOrienation Conversion Routine.
 * Returns True if Conversion is successful.
 */
Boolean
XmuCvtStringToPageOrientation(Display *dpy, XrmValue *args, Cardinal *num_args, XrmValue *fromVal, XrmValue *toVal, XtPointer *data)
{
    static XrmQuark		XrmQEportrait;
    static XrmQuark		XrmQElandscape;
    static XrmQuark		XrmQEupsideDown;
    static XrmQuark		XrmQEseascape;
    static int			haveQuarks;
    XrmQuark    q;
    char	*str = (XPointer) fromVal->addr;
    char        lowerName[1000];

    if (str == NULL) return False;

    if (!haveQuarks) {
	XrmQEportrait   = XrmStringToQuark(XtEportrait);
	XrmQElandscape  = XrmStringToQuark(XtElandscape);
	XrmQEupsideDown = XrmStringToQuark(XtEupsideDown);
	XrmQEseascape   = XrmStringToQuark(XtEseascape);
	haveQuarks = 1;
    }

    XmuCopyISOLatin1Lowered(lowerName, str);

    q = XrmStringToQuark(lowerName);

    if (q == XrmQEportrait)
	done(XtPageOrientation, XtPageOrientationPortrait);
    if (q == XrmQElandscape)
	done(XtPageOrientation, XtPageOrientationLandscape);
    if (q == XrmQEupsideDown)
	done(XtPageOrientation, XtPageOrientationUpsideDown);
    if (q == XrmQEseascape)
	done(XtPageOrientation, XtPageOrientationSeascape);

    XtDisplayStringConversionWarning(dpy, str, XtRPageOrientation);
    return False;
}

/* Palette Conversion Routine.
 * Returns True if Conversion is successful.
 */
Boolean
XmuCvtStringToPalette(Display *dpy, XrmValue *args, Cardinal *num_args, XrmValue *fromVal, XrmValue *toVal, XtPointer *data)
{
    static XrmQuark		XrmQEmonochrome;
    static XrmQuark		XrmQEgrayscale;
    static XrmQuark		XrmQEcolor;
    static int			haveQuarks;
    XrmQuark    q;
    char	*str = (XPointer) fromVal->addr;
    char        lowerName[1000];

    if (str == NULL) return False;

    if (!haveQuarks) {
	XrmQEmonochrome = XrmStringToQuark(XtEmonochrome);
	XrmQEgrayscale  = XrmStringToQuark(XtEgrayscale);
	XrmQEcolor      = XrmStringToQuark(XtEcolor);
	haveQuarks = 1;
    }

    XmuCopyISOLatin1Lowered(lowerName, str);

    q = XrmStringToQuark(lowerName);

    if (q == XrmQEmonochrome)
	done(XtPalette, XtPaletteMonochrome);
    if (q == XrmQEgrayscale)
	done(XtPalette, XtPaletteGrayscale);
    if (q == XrmQEcolor)
	done(XtPalette, XtPaletteColor);

    XtDisplayStringConversionWarning(dpy, str, XtRPalette);
    return False;
}
